{# This is the template for the navigation #}

<DIV class="nav">
    <A href="{{ to_root }}">ArcticZim</A>
    -
    <A href="{{ to_root }}/subreddits/">Subreddits</A>
    -
    <A href="{{ to_root }}/info/">Informations</A>
    {% if options.with_stats: %}
        -
        <A href="{{ to_root }}/statistics/">Statistics</A>
    {% endif %}
</DIV>
