{# This is a collection of utilty links for a comment. Expects "comment" and "to_root" argument. #}

<P class="commentmeta2"><A href="{{ to_root }}/r/{{ comment.post.subreddit.name }}/{{ comment.post.id }}/#comment-{{ comment.id }}">Permalink</A></P>
