{# This is a page for the list of subreddits. Expects "subreddit_infos", "num_pages", "cur_page", "sort" and "to_root" argument. #}
{% extends "base.html.jinja" %}

{% block title %}
    <TITLE>All Subreddits - Page {{ cur_page }} posts</TITLE>
{% endblock %}

{% block bannertext %}
    <H1 class="bannertext">Subreddits</H1>
{% endblock %}

{% block content %}
    {% with to_root=to_root, subreddit_infos=subreddit_infos %}
        {% include "subredditlist.html.jinja" %}
    {% endwith %}
    {% with to_root=to_root, num_pages=num_pages, cur_page=cur_page, prefix="subreddits" %}
        {% include "pagebuttons.html.jinja" %}
    {% endwith %}
{% endblock %}
