{# This is a collection of utilty links for a post. Expects "post" and "to_root" argument. #}

<P class="postmeta2">
    <A href="{{ to_root }}/r/{{ post.subreddit.name }}/{{ post.id }}/">{{ post.comments|length - 1 }} comments</A>
    <A href="{{ to_root }}/{{ post.id }}">permalink</A>
    <A href="https://redd.it/{{ post.id }}/">original</A>
    <A href="{{ post.url }}">original-url</A>
</P>
